from __future__ import annotations

import discord

MINUS = discord.PartialEmoji(name="minus", animated=False, id=965672202013925447)
PLAYLIST = discord.PartialEmoji(name="playlist", animated=False, id=965672202093621319)
LOOP = discord.PartialEmoji(name="loop", animated=False, id=965672202143928362)
FOLDER = discord.PartialEmoji(name="folder", animated=False, id=965672202395586601)
POWER = discord.PartialEmoji(name="power", animated=False, id=965672202395586691)
SETTINGS = discord.PartialEmoji(name="settings", animated=False, id=965672202395590696)
# noinspection SpellCheckingInspection
VOLUME_DOWN = discord.PartialEmoji(name="volumedown", animated=False, id=965672202399801374)
REPEAT = discord.PartialEmoji(name="repeat", animated=False, id=965672202412388352)
PLUS = discord.PartialEmoji(name="plus", animated=False, id=965672202416570368)
NEXT = discord.PartialEmoji(name="next", animated=False, id=965672202416570428)
HOME = discord.PartialEmoji(name="home", animated=False, id=965672202420764713)
PREVIOUS = discord.PartialEmoji(name="previous", animated=False, id=965672202424950795)
MINIMIZE = discord.PartialEmoji(name="minimize", animated=False, id=965672202424963142)
PLAY = discord.PartialEmoji(name="play", animated=False, id=965672202441723994)
EQUALIZER = discord.PartialEmoji(name="equalizer", animated=False, id=965672202454323250)
RANDOM = discord.PartialEmoji(name="random", animated=False, id=965672202458509463)
MENU = discord.PartialEmoji(name="menu", animated=False, id=965672202466910238)
PAUSE = discord.PartialEmoji(name="pause", animated=False, id=965672202466910268)
FASTFORWARD = discord.PartialEmoji(name="fastforward", animated=False, id=965672202513027122)
# noinspection SpellCheckingInspection
VOLUME_UP = discord.PartialEmoji(name="volumeup", animated=False, id=965672202517225492)
STOP = discord.PartialEmoji(name="stop", animated=False, id=965672202563362926)
MUTE = discord.PartialEmoji(name="mute", animated=False, id=965672202592747550)
ARTIST = discord.PartialEmoji(name="artist", animated=False, id=965672202592747600)
SEARCH = discord.PartialEmoji(name="search", animated=False, id=965672202596921384)
# noinspection SpellCheckingInspection
STACKED_ADD_FOLDER = discord.PartialEmoji(name="stackedaddfolder", animated=False, id=965672202785661029)
# noinspection SpellCheckingInspection
MUSICAL_NOTE = discord.PartialEmoji(name="musicalnote", animated=False, id=965674278144077824)
CHECK = discord.PartialEmoji(name="check", animated=False, id=967466875535626260)
TRASH = discord.PartialEmoji(name="trash", animated=False, id=967752655017484318)
URL = discord.PartialEmoji(name="url", animated=False, id=967753966093991968)
CLEAR = discord.PartialEmoji(name="clear", animated=False, id=967756040521252924)
DOWNLOAD = discord.PartialEmoji(name="download", animated=False, id=967760183977730079)
NAME = discord.PartialEmoji(name="name", animated=False, id=967768470781579284)
UPDATE = discord.PartialEmoji(name="update", animated=False, id=967810735851860059)
INFO = discord.PartialEmoji(name="info", animated=False, id=967827814160158820)
QUEUE = discord.PartialEmoji(name="queue", animated=False, id=967902316185415681)
DUPLICATE = discord.PartialEmoji(name="duplicate", animated=False, id=967922711026343966)
SSL = discord.PartialEmoji(name="ssl", animated=False, id=969662672570294342)
HOST = discord.PartialEmoji(name="host", animated=False, id=969675288642486363)
TIMEOUT = discord.PartialEmoji(name="timeout", animated=False, id=969676222697205810)
PORT = discord.PartialEmoji(name="port", animated=False, id=969676962824069200)
PASSWORD = discord.PartialEmoji(name="password", animated=False, id=969690513911459910)
# noinspection SpellCheckingInspection
CLOUD_SERVER = discord.PartialEmoji(name="cloudserver", animated=False, id=972827895900811337)
